# this script is created by running knitr::purl(<file>, documentation = 0) on
# the .Rnw file which has been used to the create article.

# set options
options(digits = 4, scipen = 7, width = 100,
        xtable.sanitize.text.function = function(x) gsub("\\&", "\\\\&", x),
        xtable.booktabs = TRUE,
        xtable.caption.placement = "top",
        xtable.size = "\\tiny")

# load the packages we need
library(mdgc)
library(mixedgcImp)
library(missForest, quietly = TRUE)
library(xtable)
library(foreign)
library(mfp)
library(catdata)
library(missMDA)
library(doParallel)
library(parallel)
library(survival)
library(abind)

# simulates a data set and mask some of the data.
#
# Args:
#   n: number of observations.
#   p: number of variables.
#   n_lvls: number of levels for the ordinal variables.
#   p_na: chance of a NA.
#
# Returns:
#   Simulated masked data, the true data, and true covariance matrix.
sim_dat <- function(n, p = 3L, n_lvls = 5L, p_na = .3){
  # get the covariance matrix
  Sig <- cov2cor(drop(rWishart(1L, p, diag(p))))

  # draw the observations
  truth <- matrix(rnorm(n * p), n) %*% chol(Sig)

  # determine the type
  n_rep <- floor((p + 3 - 1) / 3)
  type <- rep(1:3, each = n_rep)[1:p]
  is_con <- type == 1L
  is_bin <- type == 2L
  is_ord <- type == 3L
  col_nam <- c(outer(1:n_rep, c("C", "B", "O"),
                     function(x, y) paste0(y, x)))[1:p]

  # sample which are masked data
  is_mask <- matrix(runif(n * p) < p_na, n)

  # make sure we have no rows with all missing data
  while(any(all_nans <- rowSums(is_mask) == NCOL(is_mask)))
    is_mask[all_nans, ] <- runif(sum(all_nans) * p) < p_na

  # create observed data
  truth_obs <- data.frame(truth)
  colnames(truth_obs) <- col_nam
  truth_obs[, is_con] <- qexp(pnorm(as.matrix(truth_obs[, is_con])))

  bs_border <- 0
  truth_obs[, is_bin] <-
    truth_obs[, is_bin] > rep(bs_border, each = NROW(truth_obs))

  bs_ord <- qnorm(seq(0, 1, length.out = n_lvls + 1L))
  truth_obs[, is_ord] <- as.integer(cut(truth[, is_ord], breaks = bs_ord))
  for(i in which(is_ord)){
    truth_obs[, i] <- ordered(truth_obs[, i])
    levels(truth_obs[, i]) <-
      LETTERS[seq_len(length(unique(truth_obs[, i])))]
  }

  # mask the data
  seen_obs <- truth_obs
  seen_obs[is_mask] <- NA

  list(truth = truth, truth_obs = truth_obs, seen_obs = seen_obs,
       Sigma = Sig)
}

# thresholds the output from mdgc.
#
# Args:
#   org_data: data with missing values.
#   imputed: output from mdgc.
#
# Returns:
#   data.frame that has the highest conditional probability category for the
#   binary and multinomial variables and back transforms the conditional medians
#   for the ordinal variables.
threshold <- function(org_data, imputed){
  # checks
  stopifnot(NROW(org_data) == length(imputed),
            is.list(imputed), is.data.frame(org_data))

  # threshold
  is_cont <- which(sapply(org_data, is.numeric))
  is_bin  <- which(sapply(org_data, is.logical))
  is_ord  <- which(sapply(org_data, is.ordered))
  is_mult <- which(sapply(org_data, is.factor))
  is_mult <- setdiff(is_mult, is_ord)
  stopifnot(
    length(is_cont) + length(is_bin) + length(is_ord) + length(is_mult) ==
      NCOL(org_data))
  is_cat <- c(is_bin, is_mult)

  # creates a data.frame from a matrix or vector
  trans_to_df <- function(x){
    if(is.matrix(x))
      as.data.frame(t(x))
    else
      as.data.frame(  x )
  }

  dum <- as.data.frame(matrix(0., NROW(org_data), 0))
  out_cont <- if(length(is_cont) == 0)
    dum else trans_to_df(sapply(imputed, function(x) unlist(x[is_cont])))
  out_cat <- if(length(is_cat) == 0)
    dum else trans_to_df(sapply(imputed, function(x)
      sapply(x[is_cat], which.max)))
  out_ord <- if(length(is_ord) == 0)
    dum else trans_to_df(sapply(imputed, function(x)
      sapply(x[is_ord], function(z) min(which(cumsum(z) > .5)))))
  out <- cbind(out_cont, out_cat, out_ord)

  # set factor levels etc.
  out <- out[, order(c(is_cont, is_bin, is_mult, is_ord))]
  if(length(is_bin) > 0)
    out[, is_bin] <- out[, is_bin] > 1L
  if(length(is_ord) > 0)
    for(i in is_ord)
      out[[i]] <- ordered(
        unlist(out[[i]]), labels = levels(org_data[, i]))
  if(length(is_mult) > 0)
    for(i in is_mult)
      out[[i]] <- factor(
        unlist(out[[i]]), labels = levels(org_data[, i]))

  colnames(out) <- colnames(org_data)
  out
}

# function to estimate the model with the mdgc package.
#
# Args:
#   dat: data.frame with missing values.
#   method: method to use.
#   verbose: TRUE for output during the estimation.
#   ord_use_median: TRUE if a median should be when imputing ordinal variables.
#   lr: learning rate to use.
est_mdgc <- function(dat, method = "svrg", verbose = FALSE,
                     ord_use_median = FALSE, lr = 1e-3){
  res <- mdgc(
    dat, verbose = verbose, maxpts = 10000L, minvls = 500L,
    n_threads = 4L, maxit = 50L, use_aprx = TRUE,
    conv_crit = 1e-8, method = method, rel_eps = 1e-2,
    batch_size = 100L, iminvls = 2000L, imaxit = 20000L,
    irel_eps = 1e-3, lr = lr)
  if(ord_use_median)
    res$ximp <- threshold(dat, res$imputed)
  res
}

# function to estimate the model with the mixedgcImp package.
#
# Args:
#   dat: data.frame with missing values.
est_mixedgcImp <- function(dat){
  # prepare the data
  dat_pass <- dat
  is_cat <- sapply(dat_pass, function(x) is.logical(x) | is.ordered(x))
  nlevels <- max(sapply(dat_pass[is_cat], function(x) length(unique(x))))
  dat_pass[, is_cat] <- lapply(dat_pass[, is_cat], as.integer)

  # fit the model and format the output
  fit <- impute_mixedgc(dat_pass, eps = 1e-4, nlevels = nlevels)
  ximp <- as.data.frame(fit$Ximp)
  colnames(ximp) <- colnames(dat)
  for(i in which(sapply(dat, is.ordered)))
    ximp[, i] <- ordered(ximp[, i], labels = levels(dat[, i]))
  list(vcov = fit$R, ximp = ximp)
}

# list with the estimations functions we will use
f1 <- est_mdgc
formals(f1)$ord_use_median <- FALSE
f2 <- est_mdgc
formals(f2)$ord_use_median <- TRUE
est_funcs <- list(`RQMC (max prob.)` = f1,
                  `RQMC (median)` = f2,
                  `Zhao & Udell` = est_mixedgcImp)

# computes the classification error.
#
# Args:
#   impu_dat: data.frame or matrix with the imputed variables.
#   truth: data.frame without any missing values.
#   observed: data.frame with missing entries.
#   keep: columns to keep. All are kept if NULL.
get_classif_error <- function(impu_dat, truth, observed, keep = NULL){
  if(!is.null(keep)){
    impu_dat <- impu_dat[, keep]
    truth <- truth[, keep]
    observed <- observed[, keep]
  }

  is_cat <- sapply(truth, function(x)
    is.logical(x) || is.factor(x))
  if(!any(is_cat))
    return(numeric())

  for(i in which(sapply(truth, is.factor))){
    f <- if(is.ordered(impu_dat[[i]]))
      ordered else factor
    impu_dat[[i]] <- f(
      as.character(impu_dat[[i]]), levels = levels(truth[[i]]))
  }

  is_match <- impu_dat[, is_cat] == truth[, is_cat]
  is_match <- matrix(is_match, ncol = sum(is_cat))
  is_match[!is.na(observed[, is_cat])] <- NA_integer_
  is_match[ is.na(truth   [, is_cat])] <- NA_integer_
  setNames(1 - colMeans(is_match, na.rm = TRUE),
           colnames(truth)[is_cat])
}

# computes the RMSE for continuous variables.
#
# Args:
#   impu_dat: data.frame or matrix with the imputed variables.
#   truth: data.frame without any missing values.
#   observed: data.frame with missing entries.
#   keep: columns to keep. All are kept if NULL.
get_rmse <- function(impu_dat, truth, observed, keep = NULL){
  if(!is.null(keep)){
    impu_dat <- impu_dat[, keep]
    truth <- truth[, keep]
    observed <- observed[, keep]
  }

  is_con <- sapply(truth, is.numeric)
  if(!any(is_con))
    return(numeric())
  err <- as.matrix(impu_dat[, is_con] - truth[, is_con])
  err[!is.na(observed[, is_con])] <- NA_real_
  err[ is.na(truth   [, is_con])] <- NA_real_
  sqrt(colMeans(err^2, na.rm = TRUE))
}

# computes the SMAE metric like Zhao & Udell (2020).
#
# Args:
#   impu_dat: data.frame or matrix with the imputed variables.
#   truth: data.frame without any missing values.
#   observed: data.frame with missing entries.
#   keep: columns to keep. All are kept if NULL.
get_smae <- function(impu_dat, truth, observed, keep = NULL){
  if(!is.null(keep)){
    impu_dat <- impu_dat[, keep]
    truth <- truth[, keep]
    observed <- observed[, keep]
  }

  for(i in which(sapply(truth, is.factor))){
    f <- if(is.ordered(impu_dat[[i]]))
      ordered else factor
    impu_dat[[i]] <- f(
      as.character(impu_dat[[i]]), levels = levels(truth[[i]]))
  }

  impu_dat <- sapply(impu_dat, as.numeric)
  truth    <- sapply(truth   , as.numeric)
  observed <- sapply(observed, as.numeric)

  meds <- apply(observed, 2, median, na.rm = TRUE)
  mapply(function(true_val, impu_val, is_miss, med){
    if(!any(is_miss))
      return(0.)
    true_val <- true_val[is_miss]
    impu_val <- impu_val[is_miss]
    mean(abs(true_val - impu_val)) / mean(abs(true_val - med))
  },
  true_val = as.data.frame(truth),
  impu_val = as.data.frame(impu_dat),
  is_miss  = as.data.frame(is.na(observed) & !is.na(truth)),
  med = meds)
}

# computes the relative Frobenius norm.
#
# Args:
#   estimate: the estimated matrix.
#   truth: the true matrix.
get_rel_norm <- function(estimate, truth)
  norm(estimate - truth, "F") / norm(truth, "F")

# parameters for the simulation
n_use <- 10000L
p_use <- 60L

# the seeds we will use
seeds_100 <- c(62486992L, 45024677L, 16313594L, 39659916L, 7247040L, 4739718L, 39843186L, 43375658L, 58114253L, 79709501L, 60968806L, 18720852L, 4572847L, 31276012L, 57878725L, 40842607L, 5687996L, 97778309L, 45285061L, 55738349L, 88017586L, 19511110L, 9281117L, 48342033L, 50170161L, 43758571L, 53703241L, 83312927L, 88357248L, 40063154L, 17514409L, 85307883L, 40755179L, 15416658L, 85854344L, 20668873L, 65560171L, 22561544L, 66145932L, 74691049L, 58340521L, 65499714L, 14472951L, 4994771L, 17601296L, 505104L, 41695955L, 24065766L, 88099427L, 66058585L, 93021179L, 99622599L, 25980761L, 63625629L, 66283873L, 94004972L, 19794462L, 313107L, 91962580L, 34735127L, 65682117L, 80724806L, 37106809L, 55833563L, 75455585L, 55703669L, 32177620L, 42746657L, 7410498L, 43103550L, 33056192L, 24710463L, 34889108L, 66831503L, 29373360L, 98235602L, 90517125L, 24049057L, 36397021L, 71657403L, 24943819L, 14732364L, 4444818L, 58310700L, 53607188L, 59755675L, 33991012L, 1516165L, 47483159L, 98549386L, 56628931L, 47911468L, 63341399L, 56875559L, 87553046L, 29417025L, 60481750L, 89554875L, 60529608L, 8538494L)

# run the simulations
sim_zhao_w_all <- lapply(seeds_100, function(s){
  # check if all the files exists before simulating the data set
  get_file_name <- function(nam)
    file.path("cache/sim-Zhao-all/",
              sprintf("n%d-p%d-%s-s%d.RDS", n_use, p_use,
                      gsub("[()\ ]", "", nam), s))
  have_all <- all(sapply(get_file_name(names(est_funcs)), file.exists))

  # simulate the data
  set.seed(s)
  dat <- sim_dat(n = if(have_all) 1L else n_use, p = p_use)

  # fit the model and impute values while recording the time it takes
  res <- mapply(function(fit_func, nam){
    # the file to use
    f <- get_file_name(nam)

    if(!file.exists(f)){
      # need to fit the model
      set.seed(s)
      ti <- system.time(res <- fit_func(dat$seen_obs))

      # compute the errors
      rmse <- get_rmse(
        impu_dat = res$ximp, truth = dat$truth_obs, observed = dat$seen_obs)
      cler <- get_classif_error(
        impu_dat = res$ximp, truth = dat$truth_obs, observed = dat$seen_obs)
      smae <- get_smae(
        impu_dat = res$ximp, truth = dat$truth_obs, observed = dat$seen_obs)

      # we also keep track of the probabilities for the imputed categorical
      # variable
      is_cat <- sapply(dat$seen_obs, function(x) is.logical(x) || is.factor(x))
      impu_cate_probs <- mapply(
        function(val, is_miss) table(val[is_miss]),
        val = res$ximp[, is_cat], is_miss <- as.data.frame(
        is.na(as.matrix(dat$seen_obs[, is_cat]))))

      if(!is.null(res$imputed)){
        impu_margins <- lapply(res$imputed, `[`, is_cat)
        impu_margins <- lapply(
          names(impu_margins[[1L]]),
          function(x) rowMeans(sapply(impu_margins, `[[`, x)))
      } else
        impu_margins <- NULL

      out <- list(RMSE = rmse, `Classification error` = cler,
                  SMAE = smae, vcov = res$vcov, time = ti,
                  impu_cate_probs = impu_cate_probs,
                  impu_margins = impu_margins)
      saveRDS(out, f)
    }

    readRDS(f)
  }, fit_func = est_funcs, nam = names(est_funcs), SIMPLIFY = FALSE)

  # compute the relative error for the covariance matrix estimate
  out <- lapply(res, function(x){
    x$vcov_err <- get_rel_norm(x$vcov, dat$Sigma)
    x
  })

  # gather the results. Start with the errors
  is_bin <- sapply(dat$truth_obs, is.logical)
  is_con <- sapply(dat$truth_obs, is.numeric)
  is_ord <- sapply(dat$truth_obs, is.ordered)

  is_bin <- names(is_bin)[is_bin]
  is_con <- names(is_con)[is_con]
  is_ord <- names(is_ord)[is_ord]

  # get matrices and arrays with all the errors
  err <- rbind(sapply(out, `[[`, "RMSE"),
               sapply(out, `[[`, "Classification error"))
  smae <- sapply(out, `[[`, "SMAE")

  bin_stats <- simplify2array(list(
    `Classification error` = err[is_bin, ], SMAE = smae[is_bin, ]))
  ord_stats <- simplify2array(list(
    `Classification error` = err[is_ord, ], SMAE = smae[is_ord, ]))
  con_stats <- simplify2array(list(
    `Classification error` = err[is_con, ], SMAE = smae[is_con, ]))

  # get the relative errors for the covariance matrix
  vcov_err <- sapply(out, "[[", "vcov_err")

  # get the computation times
  comp_times <- sapply(out, "[[", "time")[1:3, ]

  # print result to the console
  message("\nBinary errors")
  message(paste0(capture.output(
    apply(bin_stats, 2:3, mean)), collapse = "\n"))
  message("Ordinal errors")
  message(paste0(capture.output(
    apply(ord_stats, 2:3, mean)), collapse = "\n"))
  message("Continuous errors")
  message(paste0(capture.output(
    apply(con_stats, 2:3, mean)), collapse = "\n"))
  message("Covariance matrix relative errors")
  message(paste0(capture.output(vcov_err), collapse = "\n"))
  message("Computation times")
  message(paste0(capture.output(comp_times), collapse = "\n"))

  message("Probabilities of imputed ordinals (from classification)")
  ords_tables <- sapply(out, function(x)
    do.call(rbind, x$impu_cate_probs[is_ord]), simplify = "array")
  message(paste0(capture.output(
    apply(ords_tables, 3, function(x) colSums(x) / sum(x))), collapse = "\n"))

  message("Probabilities of imputed ordinals")
  is_ord <- sapply(dat$truth_obs[setdiff(names(dat$truth_obs),
                                         is_con)], is.ordered)
  ords_tables <- sapply(out, function(x)
    if(is.null(x$impu_margins)) NULL else
      do.call(rbind, x$impu_margins[is_ord]), simplify = "array")
  ord_tables <- sapply(
    ords_tables, function(x) if(is.null(x)) NA_real_ else colMeans(x))
  message(paste0(capture.output(do.call(rbind, ord_tables)), collapse = "\n"))

  # return
  list(bin_stats = bin_stats, ord_stats = ord_stats, con_stats = con_stats,
       vcov_err = vcov_err, comp_times = comp_times, all_output = out)
})

# create table. First we pull together the error estimates
err_res <- sapply(sim_zhao_w_all, function(x)
  sapply(x[c("bin_stats", "ord_stats", "con_stats")], function(stats)
    simplify2array(list(means = apply(stats, 2:3, mean),
                        vars  = apply(stats, 2:3, var) / dim(stats)[1])),
    simplify = "array"), simplify = "array")
n_sim <- dim(err_res)[5]
err_res <- apply(err_res, 1:4, mean)

# re-scale the variance estimators such that they are Monte Carlo estimators
err_res[, , "vars", ] <- sqrt(err_res[, , "vars", ] / n_sim)

# ensure that the dimension are as they should
stopifnot(isTRUE(all.equal(
  dimnames(err_res),
  list(c("RQMC (max prob.)", "RQMC (median)", "Zhao & Udell"), c("Classification error", "SMAE"), c("means", "vars"), c("bin_stats", "ord_stats", "con_stats")))))

# remove redundant entries
err_res["RQMC (median)", , , c("bin_stats", "con_stats")] <- NA_real_

# format digits.
#
# Args:
#   x: matrix with each pair of columns being the estimates and the standard
#   errors.
#   grp: row group indicator.
#   digits_use: number of digits to use.
#   digits: number of expected significant digits.
format_table_cell <- function(x, grp = rep(1, NROW(x)), digits_use,
                              digits = 3L){
  n_col_out <- NCOL(x) / 2L
  out <- matrix("", NROW(x), n_col_out)
  stopifnot(length(digits_use) == n_col_out)

  for(i in 1:n_col_out){
    idx <- 1:2 + (i - 1L) * 2L

    if(all(is.na(x[, idx])))
      next

    # want +/- 2 SE
    x[, idx[2]] <- 2 * x[, idx[2]]

    # find the rows to make bold
    idx_min <- by(cbind(1:NROW(x), x[, idx[1]]), grp, function(x)
      x[which(x[, 2] <= min(x[, 2], na.rm = TRUE)), 1])

    # get the number of digits and format
    n_digits <- max(
      max(0, -floor(log10(x[, idx[1]])) + digits - 1L, na.rm = TRUE),
      -floor(log10(x[, idx[2]])), na.rm = TRUE)
    stopifnot(n_digits <= digits_use[i])
    n_digits <- digits_use[i]

    fmt <- paste0("%.", n_digits, "f")
    lhs <- ifelse(is.na(x[, idx[1]]), "", sprintf(fmt, x[, idx[1]]))

    rhs <- round(x[, idx[2]] * 10^(n_digits))
    rhs <- ifelse(is.na(x[, idx[2]]), "", as.character(rhs))
    out[, i] <- ifelse(nchar(lhs) == 0, "", paste0(lhs, "(", rhs, ")"))

    # make some entries bold
    idx_min <- unlist(idx_min)
    out[idx_min, i] <- paste("\\bftab", out[idx_min, i])
  }

  out
}

#####
# get the average computation time and relative errors for the covariance matrix
cmp_times <- sapply(sim_zhao_w_all, `[[`, "comp_times",
                    simplify = "array")["elapsed", , ]
vcov_err <- sapply(sim_zhao_w_all, `[[`, "vcov_err")

stat <- function(x)
  c(mean = mean(x), SE = sd(x) / sqrt(length(x)))
time_vcov <- simplify2array(list(
  `Computation time`          = apply(cmp_times, 1, stat),
  `Relative covariance error` = apply(vcov_err , 1, stat)))
time_vcov[, "RQMC (median)", ] <- NA_real_

# permute and add empty rows
time_vcov <- aperm(time_vcov, c(2L, 1L, 3L))
time_vcov <- apply(time_vcov, 2:3, function(x){
  c(x, rep(NA_real_, length(x)))
})
dnames <- dimnames(time_vcov)
dim(time_vcov) <- c(dim(err_res)[1:2], dim(time_vcov)[-1])
dimnames(time_vcov)[3:4] <- dnames[2:3]

# create table
tab <- abind(err_res, time_vcov)
tab <- matrix(tab, nrow = prod(dim(tab)[1:2]),
              ncol = prod(dim(tab)[3:4]))
tab <- format_table_cell(tab, grp = gl(2, 3), c(4L, 3L, 3L, 0L, 4L))

# set column names
colnames(tab) <- c("{Binary}", "{Ordinal}", "{Continuous}", "{Time}", "{Relative $\\mat\\Sigma$ error}")

# add the row labels and create the table
tab <- cbind(Metric = c("Error", "", "", "SMAE", "", ""),
             Method = rep(dimnames(err_res)[[1]], 2), tab)

alter_row_names <- Vectorize(function(x)
  switch (x,
          RQMC = "RQMC (our)",
          `RQMC (max prob.)` = "RQMC (our)",
          `RQMC (median)` = "Median (our)",
          `Zhao & Udell` = "AEM (ZU)",
          est_missForest = "missForest",
          x))
tab[, "Method"] <- alter_row_names(tab[, "Method"])

print(
  xtable(tab, caption = "Mean classification error, \\acs{RMSE}, and \\acs{SMAE} each with plus or minus two standard errors for the first simulation study. The error rows are the mean classification errors for the ordinal variables and the mean \\acsp{RMSE} of the continuous variables. The last two columns show the mean computation times in seconds and relative errors for the correlation matrix estimates. The best result in each comparison is in bold.",
         label = "tab:simZhaoError", align = c("l", "l", "l", "S[table-format = 1.4(4)]", "S[table-format = 1.3(3)]", "S[table-format = 1.3(3)]|", "S[table-format = 3(3)]", "S[table-format = 1.4(4)]")),
  include.rownames = FALSE, floating.environment = "table*",
  table.placement = "t")

# we run the simulation where we vary the number of samples in another
# environment then the global environment to avoid overwriting any parameters
vary_n_env <- new.env() # the environment we will use
local(envir = vary_n_env, {
  # change the number seeds
  seeds_use <- head(seeds_100, 50)

  # use a subset of the methods
  f1 <- est_mdgc
  formals(f1)$ord_use_median <- TRUE
  est_funcs <- list(`RQMC (median)` = f1,
                    `Zhao & Udell` = est_mixedgcImp)

  	# set the number of parameters to use and the sequence of sample sizes
    n_use <- 2^(9:16)
    p_use <- 15L

  # run the simulations
  sim_w_diff_size <- lapply(n_use, function(n) lapply(seeds_use, function(s){
    # check if all the files exists before simulating the data set
    get_file_name <- function(nam)
      file.path("cache/sim-sample-size/",
                sprintf("n%d-p%d-%s-s%d.RDS", n, p_use,
                        gsub("[()\ ]", "", nam), s))
    have_all <- all(sapply(get_file_name(names(est_funcs)), file.exists))

    # simulate the data
    set.seed(s)
    dat <- sim_dat(n = if(have_all) 1L else n, p = p_use)

    # fit the model and impute values while recording the time it takes
    res <- mapply(function(fit_func, nam){
      # the file to use
      f <- get_file_name(nam)

      if(!file.exists(f)){
        # need to fit the model
        set.seed(s)
        ti <- system.time(res <- fit_func(dat$seen_obs))

        # compute the errors
        rmse <- get_rmse(
          impu_dat = res$ximp, truth = dat$truth_obs, observed = dat$seen_obs)
        cler <- get_classif_error(
          impu_dat = res$ximp, truth = dat$truth_obs, observed = dat$seen_obs)
        smae <- get_smae(
          impu_dat = res$ximp, truth = dat$truth_obs, observed = dat$seen_obs)

        # we also keep track of the probabilities for the imputed categorical
        # variable
        is_cat <- sapply(dat$seen_obs, function(x) is.logical(x) || is.factor(x))
        impu_cate_probs <- mapply(
          function(val, is_miss) table(val[is_miss]),
          val = res$ximp[, is_cat], is_miss <- as.data.frame(
          is.na(as.matrix(dat$seen_obs[, is_cat]))))

        if(!is.null(res$imputed)){
          impu_margins <- lapply(res$imputed, `[`, is_cat)
          impu_margins <- lapply(
            names(impu_margins[[1L]]),
            function(x) rowMeans(sapply(impu_margins, `[[`, x)))
        } else
          impu_margins <- NULL

        out <- list(RMSE = rmse, `Classification error` = cler,
                    SMAE = smae, vcov = res$vcov, time = ti,
                    impu_cate_probs = impu_cate_probs,
                    impu_margins = impu_margins)
        saveRDS(out, f)
      }

      readRDS(f)
    }, fit_func = est_funcs, nam = names(est_funcs), SIMPLIFY = FALSE)

    # compute the relative error for the covariance matrix estimate
    out <- lapply(res, function(x){
      x$vcov_err <- get_rel_norm(x$vcov, dat$Sigma)
      x
    })

    # gather the results. Start with the errors
    is_bin <- sapply(dat$truth_obs, is.logical)
    is_con <- sapply(dat$truth_obs, is.numeric)
    is_ord <- sapply(dat$truth_obs, is.ordered)

    is_bin <- names(is_bin)[is_bin]
    is_con <- names(is_con)[is_con]
    is_ord <- names(is_ord)[is_ord]

    # get matrices and arrays with all the errors
    err <- rbind(sapply(out, `[[`, "RMSE"),
                 sapply(out, `[[`, "Classification error"))
    smae <- sapply(out, `[[`, "SMAE")

    bin_stats <- simplify2array(list(
      `Classification error` = err[is_bin, ], SMAE = smae[is_bin, ]))
    ord_stats <- simplify2array(list(
      `Classification error` = err[is_ord, ], SMAE = smae[is_ord, ]))
    con_stats <- simplify2array(list(
      `Classification error` = err[is_con, ], SMAE = smae[is_con, ]))

    # get the relative errors for the covariance matrix
    vcov_err <- sapply(out, "[[", "vcov_err")

    # get the computation times
    comp_times <- sapply(out, "[[", "time")[1:3, ]

    # print result to the console
    message(sprintf("\nn_use: %d, seed: %d", n, s))
    message("Binary errors")
    message(paste0(capture.output(
      apply(bin_stats, 2:3, mean)), collapse = "\n"))
    message("Ordinal errors")
    message(paste0(capture.output(
      apply(ord_stats, 2:3, mean)), collapse = "\n"))
    message("Continuous errors")
    message(paste0(capture.output(
      apply(con_stats, 2:3, mean)), collapse = "\n"))
    message("Covariance matrix relative errors")
    message(paste0(capture.output(vcov_err), collapse = "\n"))
    message("Computation times")
    message(paste0(capture.output(comp_times["elapsed", ]), collapse = "\n"))

    message("Probabilities of imputed ordinals (from classification)")
    ords_tables <- sapply(out, function(x)
      do.call(rbind, x$impu_cate_probs[is_ord]), simplify = "array")
    message(paste0(capture.output(
      apply(ords_tables, 3, function(x) colSums(x) / sum(x))), collapse = "\n"))

    message("Probabilities of imputed ordinals")
    is_ord <- sapply(dat$truth_obs[setdiff(names(dat$truth_obs),
                                           is_con)], is.ordered)
    ords_tables <- sapply(out, function(x)
      if(is.null(x$impu_margins)) NULL else
        do.call(rbind, x$impu_margins[is_ord]), simplify = "array")
    ord_tables <- sapply(
      ords_tables, function(x) if(is.null(x)) NA_real_ else colMeans(x))
    message(paste0(capture.output(do.call(rbind, ord_tables)), collapse = "\n"))

    # return
    list(bin_stats = bin_stats, ord_stats = ord_stats, con_stats = con_stats,
         vcov_err = vcov_err, comp_times = comp_times, all_output = out)
  }))

  # pull together the error estimates
  err_res <- lapply(sim_w_diff_size, function(z) sapply(z, function(x)
    sapply(x[c("bin_stats", "ord_stats", "con_stats")], function(stats)
      simplify2array(list(means = apply(stats, 2:3, mean),
                          vars  = apply(stats, 2:3, var) / dim(stats)[1])),
      simplify = "array"), simplify = "array"))

  vcov_err <-  lapply(sim_w_diff_size, function(z) t(sapply(z, `[[`, "vcov_err")))

  # create a box plot with the relative error for the covariance matrix
  get_row_names <- function(x)
     sapply(x, switch, `RQMC (median)` = "RQMC",
            `Zhao & Udell` = "AEM", stop("not implemented"))

  grp <- 0L
  vcov_err_box <- lapply(vcov_err, function(x){
    grp <<- grp + 1L
    err <- c(x)
    data.frame(
      err = err, sample_size = rep(n_use[grp], length(err)),
      type = as.factor(rep(get_row_names(colnames(x)), each = NROW(x))))
  })
  vcov_err_box <- do.call(rbind, vcov_err_box)
  x_shift <- .2
  where_at <- rep(c(-x_shift, x_shift), length(vcov_err)) +
    rep(1:length(vcov_err), each = 2)
  par(mar = c(4.5, 4.5, .5, .5), bty = "l", mfcol = c(1, 2), cex = .7)
  boxplot(err ~ type + sample_size, data = vcov_err_box, log = "y",
          pars = list(bty = "l", outpch = rep(16:17, length(vcov_err)),
                      outcex = .5, medlwd = 1.1),
          at = where_at, xlab = "Sample size",
          col = rep(c("white", "gray"), length(vcov_err)), xaxt = "n",
          ylab = expression(paste("Relative ", Sigma, " error")))
  axis(1, at = seq_along(vcov_err), labels = n_use)

  # fit line through the relative error and add the regression line
  sigma_fit <- lm(log(err) ~ type + log(sample_size) : type, vcov_err_box)
  curve({
    z <- x + x_shift
    z <- log(n_use)[1] + (x - 1) * log(2)
    exp(predict(sigma_fit, data.frame(type = "AEM", sample_size = exp(z))))},
    add = TRUE, lty = 2)

  curve({
    z <- x - x_shift
    z <- log(n_use)[1] + (z - 1) * log(2)
    exp(predict(sigma_fit, data.frame(type = "RQMC", sample_size = exp(z))))},
    add = TRUE, lty = 2, col = "darkgray")

  # add plots for the difference in errors
  grp <- 0L
  err_res_mea_diff <- lapply(err_res, function(x){
    grp <<- grp + 1L
    diffs <- t(apply(x[, "Classification error", "means", , ], 2:3, diff))
    data.frame(
      diffs = c(diffs), sample_size = rep(n_use[grp], length(diffs)),
      type = as.factor(rep(colnames(diffs), each = NROW(diffs))))
  })
  err_res_mea_diff <- do.call(rbind, err_res_mea_diff)

  x_range <- .4
  where_at <- rep(seq(-x_range, x_range, length.out = length(n_use)), 3) +
    rep(1:3, each = length(n_use))
  boxplot(diffs ~ sample_size + type, data = err_res_mea_diff,
          pars = list(bty = "l", outcex = .5, medlwd = 1.1),
          at = where_at, xlab = "Metric and sample size", xaxt = "n",
          ylab = "Difference in error")
  axis(1, at = 1:3,
       labels = sapply(levels(err_res_mea_diff$type), switch,
                       bin_stats = "Binary", con_stats = "Continuous",
                       ord_stats = "Ordinal"))
  abline(h = 0, lty = 2)
})

# remove the objects we do not need
vars_to_keep <- c(
  "seeds_100", "est_mdgc", "est_mixedgcImp", "format_table_cell",
  "get_classif_error", "get_rel_norm", "get_rmse", "get_smae", "threshold",
  "est_funcs", "is_min", "alter_row_names", "vars_to_keep", "re_scale",
  "data_sets", "vary_n_env")
rm(list = setdiff(ls(), vars_to_keep))

# re-scales continuous variables to have scale 1.
#
# Args:
#   dat: data to rescale.
re_scale <- function(dat){
  is_num <- sapply(dat, is.numeric)
  if(!any(is_num))
    return(dat)
  dat[is_num] <- lapply(dat[is_num], scale)
  dat[is_num] <- lapply(dat[is_num], c) # remove attributes
  dat
}

# performs a simulation study using a given data set.
#
# Args:
#   dat: data set to use.
#   fit_funcs: list with estimation functions to use.
#   prefix: prefix for the names of the saved results.
#   seeds: seeds to use.
#   p_na: chance of a NA.
#   do_mask: character with columns to mask. All columns are selected if this is
#   NULL.
perform_study <- function(dat, fit_funcs, prefix, seeds, p_na = .3,
                          do_mask = NULL){
  message_header <- paste0(deparse(substitute(dat)), collapse = "\n")
  dat <- re_scale(dat)

  # function to mask data
  mask_dat <- function(){
    # samples NA entries
    is_mask <- matrix(runif(NROW(dat) * NCOL(dat)) < p_na, NROW(dat))
    if(!is.null(do_mask)){
      stopifnot(length(setdiff(do_mask, colnames(dat))) == 0L)
      is_mask[, !colnames(dat) %in% do_mask] <- FALSE
    }

    # make sure we have no rows with all missing data
    while(any(all_nans <- rowSums(is_mask) == NCOL(is_mask)))
      is_mask[all_nans, ] <- runif(sum(all_nans) * NCOL(dat)) < p_na

    # mask data
    dat[is_mask] <- NA

    # remove unused levels and return
    droplevels(dat)
  }

  # perform the study
  j <- -1L
  lapply(seeds, function(s){
    # check if we have all the files we need
    get_file_nam <- function(nam)
      file.path("cache/", prefix,
                sprintf("%s-s%d.RDS", gsub("[()\ ;]", "", nam), s))
    have_all <- all(sapply(get_file_nam(names(fit_funcs)), file.exists))

    # simulate the data
    set.seed(s)
    dat_masked <- if(have_all) dat else mask_dat()

    # fit the model and impute values while recording the time it takes
    res <- mapply(function(fit_func, nam){
      # the file to use
      f <- get_file_nam(nam)

      if(!file.exists(f)){
        set.seed(s)
        ti <- system.time(res <- fit_func(dat_masked))

        # compute the errors
        rmse <- get_rmse(
          impu_dat = res$ximp, truth = dat, observed = dat_masked,
          keep = do_mask)
        cler <- get_classif_error(
          impu_dat = res$ximp, truth = dat, observed = dat_masked,
          keep = do_mask)
        smae <- get_smae(
          impu_dat = res$ximp, truth = dat, observed = dat_masked,
          keep = do_mask)

        # we also keep track of the probabilities for the imputed categorical
        # variable
        is_cat <- sapply(dat_masked, function(x) is.logical(x) || is.factor(x))
        impu_cate_probs <- mapply(
          function(val, is_miss) table(val[is_miss]),
          val = res$ximp[, is_cat], is_miss <- as.data.frame(
          is.na(as.matrix(dat_masked[, is_cat]))))

        if(!is.null(res$imputed)){
          impu_margins <- lapply(res$imputed, `[`, is_cat)
          impu_margins <- lapply(
            names(impu_margins[[1L]]),
            function(x) rowMeans(sapply(impu_margins, `[[`, x)))
        } else
          impu_margins <- NULL

        out <- list(RMSE = rmse, `Classification error` = cler,
                    SMAE = smae, vcov = res$vcov, time = ti,
                    impu_cate_probs = impu_cate_probs,
                    impu_margins = impu_margins)
        saveRDS(out, f)
      }

      readRDS(f)
    }, fit_func = fit_funcs, nam = names(fit_funcs), SIMPLIFY = FALSE)

    # gather the results. Start with the errors
    dat_missing <- if(!is.null(do_mask)) dat[do_mask] else dat
    is_bin <- sapply(dat_missing, is.logical)
    is_con <- sapply(dat_missing, is.numeric)
    is_ord <- sapply(dat_missing, is.ordered)

    is_bin <- names(is_bin)[is_bin]
    is_con <- names(is_con)[is_con]
    is_ord <- names(is_ord)[is_ord]

    # get matrices and arrays with all the errors
    gather_res <- function(what, names){
      if(length(names) < 1L)
        return(dum <- matrix(0., 0, length(res)))
      out <- sapply(res, `[[`, what)
      if(length(names) == 1L)
        out <- matrix(out, 1L, dimnames = list(names, base::names(out)))
      out
    }

    err <- rbind(gather_res("RMSE", is_con),
                 gather_res("Classification error", c(is_bin, is_ord)))
    smae <- sapply(res, `[[`, "SMAE")

    bin_stats <- simplify2array(list(
      `Classification error` = err[is_bin, ], SMAE = smae[is_bin, ]))
    ord_stats <- simplify2array(list(
      `Classification error` = err[is_ord, ], SMAE = smae[is_ord, ]))
    con_stats <- simplify2array(list(
      `Classification error` = err[is_con, ], SMAE = smae[is_con, ]))

    # get the computation times
    comp_times <- sapply(res, "[[", "time")[1:3, ]

    # print result to the console
    j <<- j + 1L
    if(j %% 10L == 0L){
      message("")
      message(sprintf("%s. Seed %d", message_header, s))

      show_stats <- function(stats){
        to_show <- if(length(dim(stats)) == 3L)
          apply(stats, 2:3, mean) else stats
        message(paste0(capture.output(to_show), collapse = "\n"))
      }

      if(length(is_bin) > 0){
        message("Binary errors")
        show_stats(bin_stats)
      }
      if(length(is_ord) > 0){
        message("Ordinal errors")
        show_stats(ord_stats)
      }
      if(length(is_con) > 0){
        message("Continuous errors")
        show_stats(con_stats)
      }

      message("Computation times")
      message(paste0(capture.output(comp_times), collapse = "\n"))
    }

    # return
    list(bin_stats = bin_stats, ord_stats = ord_stats, con_stats = con_stats,
         comp_times = comp_times, all_output = res)
  })
}

# creates table rows with the output from perform_study.
#
# Args:
#   sim_res: output from perform_study.
#   dat: data set passed to perform_study.
#   data_set_name: name of the data set.
#   keep: methods to keep.
#   do_mask: do_mask argument used for sim_res.
#   digits_use: see format_table_cell.
create_sim_study_row <- function(sim_res, dat, data_set_name, keep = TRUE,
                                 do_mask = NULL, digits_use = c(4L, 4L, 3L, 2L)){
  msg_dat <- paste0(deparse(substitute(dat)), collapse = "\n")
  if(!is.null(do_mask))
    dat <- dat[, do_mask]

  # get indices for the different types
  is_bin <- which(sapply(dat, is.logical))
  is_ord <- which(sapply(dat, is.ordered))
  is_con <- which(sapply(dat, is.numeric))

  any_bin <- length(is_bin) > 0
  any_ord <- length(is_ord) > 0
  any_con <- length(is_con) > 0

  # First we pull together the error estimates
  to_get <- c(if(any_bin) "bin_stats" else NULL,
              if(any_ord) "ord_stats" else NULL,
              if(any_con) "con_stats" else NULL)
  err_res <- sapply(sim_res, function(x)
    sapply(x[to_get], function(x)
      if(is.matrix(x)) x else apply(x, 2:3, mean),
           simplify = "array"),
    simplify = "array")

  # compute means and standard errors
  meas <- apply(err_res, 1:3, mean)
  SEs <- apply(err_res, 1:3, sd) / sqrt(dim(err_res)[4])

  # report to console
  message(msg_dat)
  message("Means")
  message(paste0(capture.output(meas), collapse = "\n"))

  # this is very hard coded
  if(any_bin || any_con){
    meas["RQMC (median)", , c("bin_stats", "con_stats")] <- NA_real_
    SEs ["RQMC (median)", , c("bin_stats", "con_stats")] <- NA_real_
  }

  # subset
  meas <- meas[keep, , , drop = FALSE]
  SEs <- SEs[keep, , , drop = FALSE]

  # create table rows
  n_mods <- dim(meas)[[1L]]
  out <- matrix(
    NA_real_, n_mods * 2L, 8L, dimnames = list(
      rep(dimnames(meas)[[1L]], 2),
      c("Binary", " ", "Ordinal", " ", "Continuous", " ",
        "Time", "")))

  if(any_bin){
    out[, 1L] <- meas[, , "bin_stats"]
    out[, 2L] <- SEs [, , "bin_stats"]
  }
  if(any_ord){
    out[, 3L] <- meas[, , "ord_stats"]
    out[, 4L] <- SEs [, , "ord_stats"]
  }
  if(any_con){
    out[, 5L] <- meas[, , "con_stats"]
    out[, 6L] <- SEs [, , "con_stats"]
  }

  # add the computation time
  comp_times <- sapply(sim_res, function(x) x$comp_times["elapsed", ],
                       simplify = "array")
  out[, 7L] <- rowMeans(comp_times)
  out[, 8L] <- apply(comp_times, 1L, sd) / sqrt(NCOL(comp_times))

  # very hard coded
  out["RQMC (median)", 7:8] <- NA_real_

  # round, format and return
  tab <- format_table_cell(out, grp = gl(2, NROW(out) / 2),
                           digits_use = digits_use)

  # very hard coded
  tab[4:6, 4] <- ""

  tab <- cbind(c(data_set_name, rep("", NROW(tab) - 1L)),
               c("Error", rep("", NROW(tab) / 2L - 1L),
                 "SMAE", rep("", NROW(tab) / 2L - 1L)),
               rep(dimnames(meas)[[1L]], 2),
               tab)
  colnames(tab) <- c(
    "Data set", "Metric", "Method",
    paste0("{", colnames(out)[seq(1L, NCOL(out), by = 2L)], "}"))
  tab
}

# load data sets and perform the study
fit_funcs <- est_funcs

data(medcare)
medcare_use <- within(medcare, {
  hosp <- ordered(hosp)
  health <- ifelse(healthpoor == 1, "poor", ifelse(
    healthexcellent == 1, "excellent", "neither"))
  health <- ordered(health, levels = c("poor", "neither", "excellent"))
  healthpoor <- NULL
  healthexcellent <- NULL
  numchron <- ordered(numchron)
  male <- male == 1
  married <- married == 1
  school <- ordered(school)
})
data_sets <- list(Medcare = medcare_use)
medcare_res <- perform_study(dat = medcare_use, fit_funcs = fit_funcs,
                             prefix = "medcare/", seeds = seeds_100)

data(rent)
rent_use <- within(rent, {
  rooms <- ordered(rooms)
  area <- NULL
  rentm <- NULL
  good <- good == 1
  best <- best == 1
  warm <- warm == 1
  tiles <- tiles == 1
  bathextra <- bathextra == 1
  kitchen <- kitchen == 1
  central <- central == 1
})
data_sets$Rent <- rent_use
rent_res <- perform_study(dat = rent_use, fit_funcs = fit_funcs,
                          prefix = "rent/", seeds = seeds_100)

esl <- read.arff(file.path(
  "data", "datasets-arie_ben_david", "ESL.arff"))
esl <- as.data.frame(lapply(esl, ordered))
data_sets$ESL <- esl
esl_res <- perform_study(dat = esl, fit_funcs = fit_funcs,
                         prefix = "esl/", seeds = seeds_100)

lev <- read.arff(file.path(
  "data", "datasets-arie_ben_david", "LEV.arff"))
lev <- as.data.frame(lapply(lev, ordered))
data_sets$LEV <- lev
lev_res <- perform_study(dat = lev, fit_funcs = fit_funcs,
                         prefix = "lev/", seeds = seeds_100)

data(GBSG)
gbsg_use <- within(GBSG, {
 htreat <- htreat == 1
 tumgrad <- ordered(tumgrad)
 menostat <- menostat == 2
 cens <- cens == 1

 age <- scale(age)
 tumsize <- scale(tumsize)
 posnodal <- scale(posnodal)
 prm <- scale(prm)
 esm <- scale(esm)
 rfst <- scale(rfst)

 id <- NULL
})
data_sets$GBSG <- gbsg_use
gbsg_res <- perform_study(dat = gbsg_use, fit_funcs = fit_funcs,
                          prefix = "gbsg/", seeds = seeds_100)

tips <- read.csv(file.path("data", "tips.csv"))
tips <- within(tips, {
  obs <- NULL
  sex <- sex == "F"
  smoker <- smoker == "Yes"
  day <- ordered(day)
  time <- time == "Day"
  size <- ordered(size)
  totbill <- scale(totbill)
  tip <- scale(tip)
})
data_sets$TIPS <- tips
tips_res <- perform_study(dat = tips, fit_funcs = fit_funcs,
                          prefix = "tips/", seeds = seeds_100)

# create the rows with the output
rent_rows <- create_sim_study_row(
  rent_res, rent_use, data_set_name = "Rent")

medcare_rows <- create_sim_study_row(
  medcare_res, medcare_use, data_set_name = "Medcare")

esl_rows <- create_sim_study_row(
  esl_res, esl, keep = TRUE,
  data_set_name = "ESL")

lev_row <- create_sim_study_row(
  lev_res, lev, keep = TRUE,
  data_set_name = "LEV")

gbsg_row <- create_sim_study_row(
  gbsg_res, gbsg_use, keep = TRUE,
  data_set_name = "GBSG")

tips_row <- create_sim_study_row(
  tips_res, tips, keep = TRUE,
  data_set_name = "TIPS")

# gather the rows and print the table
tab <- rbind(rent_rows, medcare_rows,
             esl_rows, lev_row, gbsg_row, tips_row)

tab[, "Method"] <- alter_row_names(tab[, "Method"])

add_to_row <- seq(6L, NROW(tab) - 6L, by = 6L)

align_arg <- c(
  rep("l", 4), "S[table-format = 1.4(4)]", "S[table-format = 1.4(4)]", "S[table-format = 1.3(3)]", "|S[table-format = 2.2(2)]")
print(xtable(
  tab, caption = "Mean classification error, \\acp{RMSE}, and \\acp{SMAE} (see Table \\ref{tab:simZhaoError}) with four different data sets used by \\cite{zhao20} and the two largest data sets from the catdata package. The average computation times in seconds are shown in the last column.",
  label = "tab:ZhaoCompare", align = align_arg),
  include.rownames = FALSE,
  add.to.row = list(pos = as.list(add_to_row),
                    command = rep("[.35em] ", length(add_to_row))),
  floating.environment = "table*", table.placement = "t")

# computes diffs for the text
norm_diffs <- sapply(list(
  rent = rent_res, medcare = medcare_res, esl = esl_res, lev = lev_res,
  gbsg = gbsg_res, tips = tips_res), function(x){
    res <- sapply(x, function(z)
      with(z$all_output, norm(`RQMC (max prob.)`$vcov, "F") -
             norm(`Zhao & Udell`$vcov, "F")))
    c(mean = mean(res), SE = sd(res) / sqrt(length(res)))
  })

# clean up
rm(list = setdiff(ls(), vars_to_keep))

# simulates a data set and mask some of the data.
#
# Args:
#   n: number of observations.
#   p: number of variables.
#   n_lvls: number of levels for the ordinal and multinomial variables.
#   p_na: chance of a NA.
#
# Returns:
#   Simulated masked data, the true data, and true covariance matrix.
sim_dat <- function(n, p = 4L, n_lvls = 5L, p_na = .3){
  # determine the type
  n_rep <- floor((p + 4L - 1L) / 4L)
  type <- rep(1:4, n_rep)[1:p]
  is_con  <- type == 1L
  is_bin  <- type == 2L
  is_ord  <- type == 3L
  is_mult <- type == 4L

  col_nam <- c(outer(c("C", "B", "O", "M"), 1:n_rep, paste0))[1:p]
  idx <- head(cumsum(c(1L, ifelse(type == 4, n_lvls, 1L))), -1L)

  # get the covariance matrix
  n_latent <- p + (n_lvls - 1L) * (p %/% 4)
  Psi <- matrix(0., n_latent, n_latent)
  dim_Sig <- n_latent - sum(is_mult)
  Psi[-idx[is_mult], -idx[is_mult]] <-
    cov2cor(drop(rWishart(1L, dim_Sig, diag(dim_Sig))))

  # draw the observations
  truth <- mvtnorm::rmvnorm(n, sigma = Psi)
  truth[, idx[is_mult]] <- 0

  # sample which are masked data
  is_mask <- matrix(runif(n * p) < p_na, n)

  # make sure we have no rows with all missing data
  while(any(all_nans <- rowSums(is_mask) == NCOL(is_mask)))
    is_mask[all_nans, ] <- runif(sum(all_nans) * p) < p_na

  # create observed data
  truth_obs <- lapply(type, function(i) if(i == 1L) numeric(n) else integer(n))
  truth_obs <- data.frame(truth_obs)
  colnames(truth_obs) <- col_nam

  bs_ord <- qnorm(seq(0, 1, length.out = n_lvls + 1L))
  for(i in 1:p){
    idx_i <- idx[i]
    switch(
      type[i],
      # continuous
      truth_obs[, i] <- qexp(pnorm(truth[, idx_i])),
      # binary
      truth_obs[, i] <- truth[, idx_i] > 0,
      # ordinal
      {
        truth_obs[, i] <-
          ordered(as.integer(cut(truth[, idx_i], breaks = bs_ord)))
        levels(truth_obs[, i]) <-
          LETTERS[seq_len(length(unique(truth_obs[, i])))]
      },
      # multinomial
      {
        truth_obs[, i] <- apply(
          truth[, idx_i + 1:n_lvls - 1L, drop = FALSE], 1L, which.max)
        truth_obs[, i] <- factor(truth_obs[, i], levels = 1:n_lvls,
                                 labels = paste0("T", 1:n_lvls))
      },
      stop("Type is not implemented"))
  }

  # mask the data
  seen_obs <- truth_obs
  seen_obs[is_mask] <- NA

  list(truth = truth, truth_obs = truth_obs, seen_obs = seen_obs,
       Psi = Psi)
}

# function to estimate the model with the missForest package.
#
# Args:
#   dat: data.frame with missing values.
est_missForest <- function(dat){
  # prepare the data
  dat_pass <- dat
  is_log <- sapply(dat_pass, is.logical)
  if(any(is_log))
    dat_pass[is_log] <- lapply(dat_pass[is_log], as.factor)

  # fit the model and impute the values
  sink(tempfile())
  on.exit(sink())
  res <- missForest(dat_pass, parallelize = "forests")

  # transform and return
  if(any(is_log))
    res$ximp[is_log] <- lapply(res$ximp[is_log], function(x) as.integer(x) > 1)

  list(ximp = res$ximp)
}

# function to estimate the model with the missForest package.
#
# Args:
#   dat: data.frame with missing values.
est_imputeFAMD <- function(dat){
  # prepare the data
  dat_pass <- dat
  is_log <- sapply(dat_pass, is.logical)
  if(any(is_log))
    dat_pass[is_log] <- lapply(dat_pass[is_log], as.factor)

  # fit the model and impute the values
  ncp_max <- max(1L, min(NCOL(dat_pass) - 1L, 10L))
  if(!any(sapply(dat_pass, is.numeric)))
    ncp <- estim_ncpMCA(dat_pass, ncp.max = ncp_max,
                        ncp.min = 1L, method.cv = "Kfold", nbsim = 5L,
                        verbose = FALSE)
  else
    ncp <- estim_ncpFAMD(dat_pass, ncp.max = ncp_max,
                         ncp.min = 1L, method.cv = "Kfold", nbsim = 5L,
                         verbose = FALSE)

  res <- imputeFAMD(dat_pass, ncp = ncp$ncp)

  # transform and return
  ximp <- res$completeObs
  if(any(is_log))
    ximp[is_log] <- lapply(ximp[is_log], function(x) as.integer(x) > 1)

  # use the same factor level
  for(i in which(sapply(dat, is.factor))){
    levels(ximp[[i]]) <- levels(dat[[i]])
    if(is.ordered(dat[[i]]))
      ximp[[i]] <- as.ordered(ximp[[i]])
  }

  list(ximp = ximp, ncp = ncp)
}

# assign the estimation functions to use
f1 <- est_mdgc
formals(f1)$ord_use_median <- FALSE
est_funcs <- list(RQMC = f1, est_missForest = est_missForest,
                  imputeFAMD = est_imputeFAMD)

# parameters for the simulation
n_use <- 2000L
p_use <- 8L

# run the simulations
cl <- makeCluster(4)
registerDoParallel(cl)
sim_mult <- lapply(seeds_100, function(s){
  # check if all the files exists before simulating the data set
  get_file_name <- function(nam)
    file.path("cache/sim-mult/",
              sprintf("n%d-p%d-%s-s%d.RDS", n_use, p_use,
                      gsub("[()\ ]", "", nam), s))
  have_all <- all(sapply(get_file_name(names(est_funcs)), file.exists))

  # simulate the data
  set.seed(s)
  dat <- sim_dat(n = if(have_all) 1L else n_use, p = p_use)

  # fit the model and impute values while recording the time it takes
  res <- mapply(function(fit_func, nam){
    # the file to use
    f <- get_file_name(nam)

    if(!file.exists(f)){
      set.seed(s)
      ti <- system.time(res <- fit_func(dat$seen_obs))

      # compute the errors
      rmse <- get_rmse(
        impu_dat = res$ximp, truth = dat$truth_obs, observed = dat$seen_obs)
      cler <- get_classif_error(
        impu_dat = res$ximp, truth = dat$truth_obs, observed = dat$seen_obs)

      # we also keep track of the probabilities for the imputed categorical
      # variable
      is_cat <- sapply(dat$seen_obs, function(x) is.logical(x) || is.factor(x))
      impu_cate_probs <- mapply(
        function(val, is_miss) table(val[is_miss]),
        val = res$ximp[, is_cat], is_miss <- as.data.frame(
        is.na(as.matrix(dat$seen_obs[, is_cat]))))

      if(!is.null(res$imputed)){
        impu_margins <- lapply(res$imputed, `[`, is_cat)
        impu_margins <- lapply(
          names(impu_margins[[1L]]),
          function(x) rowMeans(sapply(impu_margins, `[[`, x)))
      } else
        impu_margins <- NULL

      out <- list(RMSE = rmse, `Classification error` = cler,
                  vcov = res$vcov, time = ti,
                  impu_cate_probs = impu_cate_probs,
                  impu_margins = impu_margins,
                  mean = res$mea, ncp = res$ncp)
      saveRDS(out, f)
    }

    readRDS(f)
  }, fit_func = est_funcs, nam = names(est_funcs), SIMPLIFY = FALSE)

  # gather the results. Start with the errors
  is_bin <- sapply(dat$truth_obs, is.logical)
  is_con <- sapply(dat$truth_obs, is.numeric)
  is_ord <- sapply(dat$truth_obs, is.ordered)
  is_fac <- sapply(dat$truth_obs, is.factor)
  is_fac <- is_fac & is_fac != is_ord

  is_bin <- names(is_bin)[is_bin]
  is_con <- names(is_con)[is_con]
  is_ord <- names(is_ord)[is_ord]
  is_fac <- names(is_fac)[is_fac]

  # get matrices and arrays with all the errors
  err <- rbind(sapply(res, `[[`, "RMSE"),
               sapply(res, `[[`, "Classification error"))

  bin_stats <- err[is_bin, ]
  ord_stats <- err[is_ord, ]
  con_stats <- err[is_con, ]
  fac_stats <- err[is_fac, ]

  # get the computation times
  comp_times <- sapply(res, "[[", "time")[1:3, ]

  # print result to the console
  message("\nBinary errors")
  message(paste0(capture.output(colMeans(bin_stats)), collapse = "\n"))
  message("Ordinal errors")
  message(paste0(capture.output(colMeans(ord_stats)), collapse = "\n"))
  message("Continuous errors")
  message(paste0(capture.output(colMeans(con_stats)), collapse = "\n"))
  message("Multinomial errors")
  message(paste0(capture.output(colMeans(fac_stats)), collapse = "\n"))
  message("Computation times")
  message(paste0(capture.output(comp_times), collapse = "\n"))

  message("ncp")
  message(paste0(capture.output(sapply(res, function(x)
    if(is.null(x$ncp)) NA_integer_ else x$ncp$ncp)), collapse = "\n"))

  message("Probabilities of imputed ordinals (from classification)")
  ords_tables <- sapply(res, function(x)
    do.call(rbind, x$impu_cate_probs[is_ord]), simplify = "array")
  message(paste0(capture.output(
    apply(ords_tables, 3, function(x) colSums(x) / sum(x))), collapse = "\n"))

  message("Probabilities of imputed ordinals")
  is_ord <- sapply(dat$truth_obs[setdiff(names(dat$truth_obs),
                                         is_con)], is.ordered)
  ords_tables <- sapply(res, function(x)
    if(is.null(x$impu_margins)) NULL else
      do.call(rbind, x$impu_margins[is_ord]), simplify = "array")
  ord_tables <- sapply(
    ords_tables, function(x) if(is.null(x)) NA_real_ else colMeans(x))
  message(paste0(capture.output(do.call(rbind, ord_tables)), collapse = "\n"))

  # return
  list(bin_stats = bin_stats, ord_stats = ord_stats, con_stats = con_stats,
       fac_stats = fac_stats, comp_times = comp_times, all_output = res)
})
stopCluster(cl)

# performs a simulation study using a given data set.
#
# Args:
#   dat: data set to use.
#   fit_funcs: list with estimation functions to use.
#   prefix: prefix for the names of the saved results.
#   seeds: seeds to use.
#   p_na: chance of a NA.
#   do_mask: character with columns to mask. All columns are selected if this is
#   NULL.
perform_study <- function(dat, fit_funcs, prefix, seeds, p_na = .3,
                          do_mask = NULL){
  message_header <- paste0(deparse(substitute(dat)), collapse = "\n")
  dat <- re_scale(dat)

  # function to mask data
  mask_dat <- function(){
    # samples NA entries
    is_mask <- matrix(runif(NROW(dat) * NCOL(dat)) < p_na, NROW(dat))
    if(!is.null(do_mask)){
      stopifnot(length(setdiff(do_mask, colnames(dat))) == 0L)
      is_mask[, !colnames(dat) %in% do_mask] <- FALSE
    }

    # make sure we have no rows with all missing data
    while(any(all_nans <- rowSums(is_mask) == NCOL(is_mask)))
      is_mask[all_nans, ] <- runif(sum(all_nans) * NCOL(dat)) < p_na

    # mask data
    dat[is_mask] <- NA

    # remove unused levels and return
    droplevels(dat)
  }

  # setup cluster to use
  cl <- makeCluster(4)
  registerDoParallel(cl)
  on.exit(stopCluster(cl))

  # perform the study
  j <- -1L
  lapply(seeds, function(s){
    # check if we have all the files we need
    get_file_nam <- function(nam)
      file.path("cache/", prefix,
                sprintf("%s-s%d.RDS", gsub("[()\ ;]", "", nam), s))
    have_all <- all(sapply(get_file_nam(names(fit_funcs)), file.exists))

    # simulate the data
    set.seed(s)
    dat_masked <- if(have_all) dat else mask_dat()

    # fit the model and impute values while recording the time it takes
    res <- mapply(function(fit_func, nam){
      # the file to use
      f <- get_file_nam(nam)

      if(!file.exists(f)){
        set.seed(s)
        ti <- system.time(res <- fit_func(dat_masked))

        # compute the errors
        rmse <- get_rmse(
          impu_dat = res$ximp, truth = dat, observed = dat_masked,
          keep = do_mask)
        cler <- get_classif_error(
          impu_dat = res$ximp, truth = dat, observed = dat_masked,
          keep = do_mask)

        # we also keep track of the probabilities for the imputed categorical
        # variable
        is_cat <- sapply(dat_masked, function(x) is.logical(x) || is.factor(x))
        impu_cate_probs <- mapply(
          function(val, is_miss) table(val[is_miss]),
          val = res$ximp[, is_cat], is_miss <- as.data.frame(
          is.na(as.matrix(dat_masked[, is_cat]))))

        if(!is.null(res$imputed)){
          impu_margins <- lapply(res$imputed, `[`, is_cat)
          impu_margins <- lapply(
            names(impu_margins[[1L]]),
            function(x) rowMeans(sapply(impu_margins, `[[`, x)))
        } else
          impu_margins <- NULL

        out <- list(RMSE = rmse, `Classification error` = cler,
                    vcov = res$vcov, time = ti,
                    impu_cate_probs = impu_cate_probs,
                    impu_margins = impu_margins,
                    ncp = res$ncp)
        saveRDS(out, f)
      }

      readRDS(f)
    }, fit_func = fit_funcs, nam = names(fit_funcs), SIMPLIFY = FALSE)

    # gather the results. Start with the errors
    dat_missing <-
      if(!is.null(do_mask)) dat[do_mask] else dat

    is_bin <- sapply(dat_missing, is.logical)
    is_con <- sapply(dat_missing, is.numeric)
    is_ord <- sapply(dat_missing, is.ordered)
    is_fac <- sapply(dat_missing, is.factor)
    is_fac <- is_fac != is_ord & is_fac

    is_bin <- names(is_bin)[is_bin]
    is_con <- names(is_con)[is_con]
    is_ord <- names(is_ord)[is_ord]
    is_fac <- names(is_fac)[is_fac]

    # get matrices and arrays with all the errors
    gather_res <- function(what, names){
      if(length(names) < 1L)
        return(dum <- matrix(0., 0, length(res)))
      out <- sapply(res, `[[`, what)
      if(length(names) == 1L)
        out <- matrix(out, 1L, dimnames = list(names, base::names(out)))
      out
    }

    err <- rbind(gather_res("RMSE", is_con),
                 gather_res("Classification error", c(is_bin, is_ord, is_fac)))

    bin_stats <- err[is_bin, ]
    ord_stats <- err[is_ord, ]
    con_stats <- err[is_con, ]
    fac_stats <- err[is_fac, ]

    # get the computation times
    comp_times <- sapply(res, "[[", "time")[1:3, ]

    # print result to the console
    j <<- j + 1L
    if(j %% 10L == 0L){
      message("")
      message(sprintf("%s. Seed %d", message_header, s))

      show_stats <- function(stats){
        to_show <- if(length(dim(stats)) == 2L)
          colMeans(stats) else stats
        message(paste0(capture.output(to_show), collapse = "\n"))
      }

      if(length(is_bin) > 0){
        message("Binary errors")
        show_stats(bin_stats)
      }
      if(length(is_ord) > 0){
        message("Ordinal errors")
        show_stats(ord_stats)
      }
      if(length(is_con) > 0){
        message("Continuous errors")
        show_stats(con_stats)
      }
      if(length(is_fac) > 0){
        message("Multinomial errors")
        show_stats(fac_stats)
      }

      message("ncp")
      message(paste0(capture.output(sapply(res, function(x)
        if(is.null(x$ncp)) NA_integer_ else x$ncp$ncp)), collapse = "\n"))

      message("Computation times")
      message(paste0(capture.output(comp_times), collapse = "\n"))
    }

    # return
    list(bin_stats = bin_stats, ord_stats = ord_stats, con_stats = con_stats,
         fac_stats = fac_stats, comp_times = comp_times, all_output = res)
  })
}

# creates table rows with the output from perform_study.
#
# Args:
#   sim_res: output from perform_study.
#   dat: data set passed to perform_study.
#   data_set_name: name of the data set.
#   keep: methods to keep.
#   do_mask: do_mask argument used for sim_res.
#   digits_use: see format_table_cell.
create_sim_study_row <- function(sim_res, dat, data_set_name, keep = TRUE,
                                 do_mask = NULL, digits_use = c(4L, 3L, 3L, 3L, 2L)){
  msg_dat <- paste0(deparse(substitute(dat)), collapse = "\n")
  if(!is.null(do_mask))
    dat <- dat[, do_mask]

  # get indices for the different types
  is_bin <- which(sapply(dat, is.logical))
  is_ord <- which(sapply(dat, is.ordered))
  is_con <- which(sapply(dat, is.numeric))
  is_fac <- which(sapply(dat, is.factor))
  is_fac <- setdiff(is_fac, is_ord)

  any_bin <- length(is_bin) > 0
  any_ord <- length(is_ord) > 0
  any_con <- length(is_con) > 0
  any_fac <- length(is_fac) > 0

  # First we pull together the error estimates
  to_get <- c(if(any_bin) "bin_stats" else NULL,
              if(any_ord) "ord_stats" else NULL,
              if(any_con) "con_stats" else NULL,
              if(any_fac) "fac_stats" else NULL)
  err_res <- sapply(sim_res, function(x)
    sapply(x[to_get], function(z)
      if(is.null(dim(z))) z else colMeans(z),
           simplify = "array"),
    simplify = "array")

  # compute means and standard errors
  meas <- apply(err_res, 1:2, mean)
  SEs <- apply(err_res, 1:2, sd) / sqrt(dim(err_res)[3])

  # report to console
  message(msg_dat)
  message("Means")
  message(paste0(capture.output(meas), collapse = "\n"))

  # subset
  meas <- meas[keep, , drop = FALSE]
  SEs <- SEs[keep, , drop = FALSE]

  # create table rows
  n_mods <- dim(meas)[[1L]]
  out <- matrix(
    NA_real_, n_mods, 10L, dimnames = list(
      dimnames(meas)[[1L]],
      c("{Binary}", " ", "{Ordinal}", " ", "{Continuous}", " ",
        "{Multinomial}", "", "{Time}", "")))

  if(any_bin){
    out[, 1L] <- meas[, "bin_stats"]
    out[, 2L] <- SEs [, "bin_stats"]
  }
  if(any_ord){
    out[, 3L] <- meas[, "ord_stats"]
    out[, 4L] <- SEs [, "ord_stats"]
  }
  if(any_con){
    out[, 5L] <- meas[, "con_stats"]
    out[, 6L] <- SEs [, "con_stats"]
  }
  if(any_fac){
    out[, 7L] <- meas[, "fac_stats"]
    out[, 8L] <- SEs [, "fac_stats"]
  }

  # add the computation time
  comp_times <- sapply(sim_res, function(x) x$comp_times["elapsed", ],
                       simplify = "array")
  out[,  9L] <- rowMeans(comp_times)
  out[, 10L] <- apply(comp_times, 1L, sd) / sqrt(NCOL(comp_times))

  # round, format and return
  tab <- format_table_cell(out, digits_use = digits_use)

  tab <- cbind(c(data_set_name, rep("", NROW(tab) - 1L)),
               dimnames(meas)[[1L]],
               tab)
  colnames(tab) <- c("Data set", "Method",
                     colnames(out)[seq(1L, NCOL(out), by = 2L)])
  tab
}

# perform the simulation study
data("nhanes", package = "survey")

nhanes_use <- within(nhanes, {
  HI_CHOL <- HI_CHOL > 0
  race <- factor(race)
  agecat <- ordered(agecat)
  RIAGENDR <- RIAGENDR > 1
})[, c("HI_CHOL", "race", "agecat", "RIAGENDR")]
nhanes_use <- nhanes_use[complete.cases(nhanes_use), ]
data_sets$Cholesterol <- nhanes_use
nhanes_res <- perform_study(
  dat = nhanes_use, fit_funcs = est_funcs, prefix = "nhanes/",
  seeds = seeds_100)

data(rent)
rent_use <- within(rent, {
  rooms <- ordered(rooms)
  area <- factor(area)
  rentm <- NULL
  good <- good == 1
  best <- best == 1
  warm <- warm == 1
  tiles <- tiles == 1
  bathextra <- bathextra == 1
  kitchen <- kitchen == 1
  central <- central == 1
})
data_sets$Rent <- rent_use
rent_res <- perform_study(dat = rent_use, fit_funcs = est_funcs,
                          prefix = "rent-mult/", seeds = seeds_100)

colon_use <- colon[, setdiff(
  colnames(colon), c("id", "study", "time", "status", "node4", "etype"))]
colon_use <- within(colon_use, {
  sex <- sex > 0
  obstruct <- obstruct > 0
  perfor <- perfor > 0
  adhere <- adhere > 0
  differ <- ordered(differ)
  extent <- ordered(extent)
  surg <- surg > 0
})
colon_use <- colon_use[complete.cases(colon_use), ]
data_sets$Colon <- colon_use
colon_res <- perform_study(dat = colon_use, fit_funcs = est_funcs,
                           prefix = "colon/", seeds = seeds_100)

# prepare table rows
sim_row <- create_sim_study_row(sim_res = sim_mult,
                                dat = sim_dat(1L, p_use)$seen_obs,
                                data_set_name = "Simulation")

nhanes_row <- create_sim_study_row(sim_res = nhanes_res, dat = nhanes_use,
                                   data_set_name = "Cholesterol")

rent_row <- create_sim_study_row(sim_res = rent_res, dat = rent_use,
                                 data_set_name = "Rent")

colon_row <- create_sim_study_row(sim_res = colon_res, dat = colon_use,
                                  data_set_name = "Colon")

# create the table
tab <- rbind(sim_row, nhanes_row, rent_row, colon_row)
tab[, "Method"] <- alter_row_names(tab[, "Method"])

add_to_row <- seq(3L, NROW(tab) - 3L, by = 3L)

align_arg <- c(
  "l", "l", "l", "S[table-format = 1.4(4)]", "S[table-format = 1.3(3)]",
  "S[table-format = 1.3(3)]", "S[table-format = 1.3(3)]", "|S[table-format = 2.2(2)]")
print(xtable(
  tab, caption = "Average \\acp{RMSE} and classification errors on simulated data sets and three observational data sets. The average computation times in seconds are given in the last column.",
  label = "tab:multRes", align = align_arg),
  include.rownames = FALSE,
  add.to.row = list(pos = as.list(add_to_row),
                    command = rep("[.35em] ", length(add_to_row))),
  floating.environment = "table*", table.placement = "t")

# get summary figures for data sets
sum_stat_tbl <- t(sapply(data_sets, function(x){
  is_num <- sapply(x, is.numeric)
  is_log <- sapply(x, is.logical)
  is_ord <- sapply(x, is.ordered)
  is_fac <- sapply(x, function(z) is.factor(z) && !is.ordered(z))

  max_ord <- sapply(x[is_ord], function(z) length(levels(z)))
  max_fac <- sapply(x[is_fac], function(z) length(levels(z)))

  if(length(max_ord) < 1)
    max_ord <- 0L
  if(length(max_fac) < 1)
    max_fac <- 0L

  c(`\\# Observations` = NROW(x),
    Binary = sum(is_log), Continuous = sum(is_num),
    Ordinal = sum(is_ord), Levels = max(max_ord),
    Multinomial = sum(is_fac), Levels = max(max_fac))
}))

print(xtable(
  sum_stat_tbl, caption = "Summary statistics for each observational data set. The first column is the number observations. The other columns indicates the number of variables of each type except for the ``levels'' columns which show the maximum number of categories for the ordinal and multinomial variables, respectively.",
  label = "tab:dataStats", align = "lrrrrrrr"),
  floating.environment = "table*", table.placement = "t")
